class SchemaUpdateConflict:
    CONFLICT_TYPES = {
        'required_no_default':  ("Required field '{}' added "
                                 "without supplying a default value"),
        'incompatible_type':    "Incompatible change in type of field '{}'",
        'incompatible_choices': ("Choices list for field '{}' doesn't "
                                 "include current field value")
    }

    def __init__(self, field, typ, **kwargs):
        if typ not in self.CONFLICT_TYPES:
            raise ValueError(
                "Invalid type '{}' for SchemaUpdateConflict".format(typ)
            )
        self.field = field
        self.typ = typ
        self.args = kwargs

    def message(self):
        return self.CONFLICT_TYPES[self.typ].format(self.field)

    def __eq__(self, other):
        return self.field == other.field and self.typ == other.typ


class SchemaUpdateException(Exception):
    def __init__(self, *args, **kwargs):
        self.conflicts = kwargs.pop('conflicts', [])
        super().__init__(*args, **kwargs)

    @property
    def messages(self):
        if not hasattr(self, '_messages'):
            self._messages = list(map(lambda c: c.message(), self.conflicts))
        return self._messages

    def __str__(self):
        return 'SchemaUpdateException: ' + str(self.messages)

    def __repr__(self):
        return str(self)



from collections import UserDict
import json
from datetime import date, datetime

from psycopg2.extras import Json

from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import JSONField

from .models import Schema, compose_schemas
from .exceptions import SchemaUpdateConflict, SchemaUpdateException


class JSONAttributes(UserDict):
    def __init__(self, data={}, *args, **kwargs):
        self._init_done = False
        self._schemas = None
        self._instance = None
        self._setup = False
        self._saved_selectors = None
        self._db_val = data
        super().__init__(data, *args, **kwargs)
        self._init_done = True

    def setup_from_dict(self, dict):
        self.setup_schema()
        if dict is None or len(dict) == 0:
            self._setup = False
        else:
            self._check_required_keys(dict.keys())
            for k, v in dict.items():
                self._check_key(k)
                self._attrs[k].validate(v)
                self[k] = v

    def setup_schema(self, schemas=None):
        if self._setup and schemas is None:
            return

        # Determine schemas for model instance containing this field.
        if schemas is not None:
            self._schemas = schemas
        else:
            self._schemas = Schema.objects.from_instance(self._instance)
        self._setup = True

        # Extract schema attributes, names of required attributes and
        # names of attributes with defaults, composing schemas for
        # instance.
        attrs = compose_schemas(*self._schemas)
        self._attrs, self._required_attrs, self._default_attrs = attrs

        # Fill in defaulted attributes.
        if len(self._required_attrs) > 0:
            for key in self._required_attrs:
                if (self._attrs[key].default is not None and
                   self._attrs[key].default != ''):
                    self[key] = self._attrs[key].default

    def _pre_save_selector_check(self, strict=False):
        if not self._setup:
            self.setup_from_dict(self._db_val)
        old_selectors = self._saved_selectors
        new_selectors = Schema.objects._get_selectors(self._instance)
        self._saved_selectors = new_selectors
        if (old_selectors is None or
           not any([n != o for n, o in zip(old_selectors, new_selectors)])):
            return
        self._setup = False
        schemas_s = self._schemas
        attrs_s = self._attrs
        required_attrs_s = self._required_attrs
        default_attrs_s = self._default_attrs
        self.setup_schema()
        conflicts = self._attr_list_conflicts(attrs_s, self._attrs,
                                              strict=strict)
        if conflicts is not None and len(conflicts) > 0:
            self._schemas = schemas_s
            self._attrs = attrs_s
            self._required_attrs = required_attrs_s
            self._default_attrs = default_attrs_s
            raise SchemaUpdateException(conflicts=conflicts)

    def _attr_list_conflicts(self, old_attrs, new_attrs, strict=False):
        conflicts = []
        for aname, a in new_attrs.items():
            val = super().get(aname, None)
            if aname not in old_attrs:
                if a.required and (a.default is None or len(a.default) == 0):
                    conflicts.append(
                        SchemaUpdateConflict(aname, 'required_no_default')
                    )
            else:
                olda = old_attrs[aname]
                if not types_compatible(a.attr_type, olda.attr_type, val):
                    conflicts.append(
                        SchemaUpdateConflict(aname, 'incompatible_type')
                    )
                if (strict and
                   not choices_compatible(a.choices, olda.choices, val)):
                    conflicts.append(
                        SchemaUpdateConflict(aname, 'incompatible_choices')
                    )
        return conflicts

    def _check_required_keys(self, keys):
        for req in self._required_attrs:
            if req not in keys and req not in self._default_attrs:
                raise ValidationError(
                    _('Missing required field %(field)s'),
                    params={'field': req}
                )

    def _check_key(self, key):
        self.setup_schema()
        if key not in self._attrs and key not in self:
            raise KeyError(key)

    def __getitem__(self, key):
        if not self._setup:
            self.setup_from_dict(self._db_val)
        self._check_key(key)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        if self._init_done:
            self._check_key(key)
            self._attrs[key].validate(value)
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        self._check_key(key)
        if key in self._required_attrs:
            raise KeyError(key)
        return super().__delitem__(key)

    @property
    def schemas(self):
        self.setup_schema()
        return self._schemas

    @property
    def attributes(self):
        self.setup_schema()
        return self._attrs


def types_compatible(new_type, old_type, value):
    return (new_type == old_type or
            new_type.name == 'text' or new_type.name == 'text_multiline')


def choices_compatible(new_choices, old_choices, value):
    return new_choices is None or len(new_choices) == 0 or value in new_choices


def schema_update_conflicts(instance):
    if not hasattr(instance, '_attr_field'):
        raise ValueError("instance doesn't have an attribute field")
    conflicts = []
    try:
        instance._attr_field._pre_save_selector_check(strict=True)
    except SchemaUpdateException as exc_info:
        conflicts = exc_info.conflicts
    return conflicts


# This is needed to provide JSON serialisation for date objects
# whenever they're saved to JSON attribute fields.  This function is
# passed as the custom "dumps" method for psycopg2's Json class to
# use.

def json_serialiser(obj):
    return json.dumps(
        obj,
        default=lambda obj: (
            obj.isoformat()
            if isinstance(obj, datetime) or isinstance(obj, date)
            else None
        )
    )


class JSONAttributeField(JSONField):
    description = _('A managed JSON attribute set')

    def __init__(self, *args, **kwargs):
        kwargs['default'] = JSONAttributes
        super().__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection, context):
        return value

    def get_prep_value(self, value):
        if value is None:
            return None
        return Json(dict(value), dumps=json_serialiser)

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type in ('has_key', 'has_keys', 'has_any_keys'):
            return value
        return (Json(dict(value), dumps=json_serialiser)
                if isinstance(value, dict)
                else super().get_prep_lookup(lookup_type, value))

    # def validate(self, value, model_instance):
    #     super(JSONField, self).validate(value, model_instance)
    #     try:
    #         json.dumps(dict(value))
    #     except TypeError:
    #         raise exceptions.ValidationError(
    #             self.error_messages['invalid'],
    #             code='invalid',
    #             params={'value': value},
    #         )



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-10-06 15:54
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('jsonattrs', '0004_convert_multilingual_fields'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='attribute',
            name='choice_labels',
        ),
        migrations.RemoveField(
            model_name='attribute',
            name='long_name',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-10-06 15:54
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.migrations.operations.special


class Migration(migrations.Migration):

    dependencies = [
        ('jsonattrs', '0002_attribute_choice_labels'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='choice_labels_xlat',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='attribute',
            name='long_name_xlat',
            field=django.contrib.postgres.fields.jsonb.JSONField(default={}),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='schema',
            name='default_language',
            field=models.CharField(blank=True, max_length=3),
        ),
    ]



from django.core.management.base import BaseCommand

from jsonattrs.models import AttributeType, create_attribute_types


def run(force=False):
    if force:
        AttributeType.objects.all().delete()
    create_attribute_types()


class Command(BaseCommand):
    help = "Load default attribute types."

    def add_arguments(self, parser):
        parser.add_argument(
            '--force',
            action='store_true',
            dest='force',
            default=False,
            help='Force object deletion and recreation'
        )

    def handle(self, *args, **options):
        run(force=options['force'])



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-09-23 07:28
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jsonattrs', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='choice_labels',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=512), null=True, size=None),
        ),
    ]



from django.db.models.signals import post_init, pre_save
from django.core.exceptions import FieldError

from .fields import JSONAttributes, JSONAttributeField
from .signals import attribute_model_pre_save


def fixup_instance(sender, **kwargs):
    instance = kwargs['instance']
    for f in instance._meta.fields:
        if isinstance(f, JSONAttributeField):
            fi = getattr(instance, f.name)
            if not isinstance(fi, JSONAttributes):
                setattr(instance, f.name, JSONAttributes(fi))
            fld = getattr(instance, f.name)
            fld._instance = instance
            if hasattr(instance, '_attr_field'):
                raise FieldError('multiple JSONAttributeField fields: '
                                 'only one is allowed per model!')
            instance._attr_field = fld
    if not hasattr(instance, '_attr_field'):
        raise FieldError('missing JSONAttributeField field in '
                         'fixup_instance decorator')


def fix_model_for_attributes(cls):
    post_init.connect(fixup_instance, sender=cls)
    pre_save.connect(attribute_model_pre_save, sender=cls)
    return cls



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-10-06 15:54
from __future__ import unicode_literals

from django.db import migrations


def convert_attribute_long_names(apps, schema_editor):
    AttributeType = apps.get_model('jsonattrs', 'AttributeType')
    Attribute = apps.get_model('jsonattrs', 'Attribute')

    # Empty database?  Probably during testing...
    if not AttributeType.objects.exists():
        return

    for attr in Attribute.objects.all():
        attr.long_name_xlat = attr.long_name
        attr.save()


def convert_choice_labels(apps, schema_editor):
    AttributeType = apps.get_model('jsonattrs', 'AttributeType')
    Attribute = apps.get_model('jsonattrs', 'Attribute')

    # Empty database?  Probably during testing...
    if not AttributeType.objects.exists():
        return

    for attr in Attribute.objects.all():
        attr.choice_labels_xlat = attr.choice_labels
        attr.save()


class Migration(migrations.Migration):

    dependencies = [
        ('jsonattrs', '0003_prepare_multilingual_setup'),
    ]

    operations = [
        migrations.RunPython(
            code=convert_attribute_long_names,
            reverse_code=migrations.operations.special.RunPython.noop,
        ),
        migrations.RunPython(
            code=convert_choice_labels,
            reverse_code=migrations.operations.special.RunPython.noop,
        ),
    ]
